part of '/masamune_location.dart';

/// Class for handling terminal location information.
///
/// The acquisition is initiated by doing [listen].
///
/// It monitors the data at each interval and calls [notifyListeners] if it has been updated.
///
/// [LocationData] is stored in [value], so refer to it.
///
/// Terminate acquisition with [unlisten].
///
/// 端末の位置情報を取り扱うクラス。
///
/// [listen]を行うことで取得を開始します。
///
/// インターバルごとにデータを監視し更新されていれば[notifyListeners]を呼び出します。
///
/// [value]に[LocationData]が格納されるのでそこを参照してください。
///
/// [unlisten]で取得を終了します。
class Location
    extends MasamuneControllerBase<LocationData?, LocationMasamuneAdapter> {
  /// Class for handling terminal location information.
  ///
  /// The acquisition is initiated by doing [listen].
  ///
  /// It monitors the data at each interval and calls [notifyListeners] if it has been updated.
  ///
  /// [LocationData] is stored in [value], so refer to it.
  ///
  /// Terminate acquisition with [unlisten].
  ///
  /// 端末の位置情報を取り扱うクラス。
  ///
  /// [listen]を行うことで取得を開始します。
  ///
  /// インターバルごとにデータを監視し更新されていれば[notifyListeners]を呼び出します。
  ///
  /// [value]に[LocationData]が格納されるのでそこを参照してください。
  ///
  /// [unlisten]で取得を終了します。
  Location({super.adapter});

  /// Query for Location.
  ///
  /// ```dart
  /// appRef.controller(Location.query(parameters));     // Get from application scope.
  /// ref.app.controller(Location.query(parameters));    // Watch at application scope.
  /// ref.page.controller(Location.query(parameters));   // Watch at page scope.
  /// ```
  static const query = _$LocationQuery();

  @override
  LocationMasamuneAdapter get primaryAdapter => LocationMasamuneAdapter.primary;

  /// Returns `true` if initialized by executing [initialize].
  ///
  /// [initialize]を実行して初期化されている場合は`true`を返します。
  bool get initialized => _initialized;
  bool _initialized = false;

  @override
  LocationData? get value => _value;
  LocationData? _value;

  final location.Location _location = location.Location();

  Timer? _timer;
  Duration _updateInterval = const Duration(seconds: 1);
  bool _updated = false;
  Completer<void>? _listenCompleter;
  Completer<void>? _initializeCompleter;
  // ignore: cancel_subscriptions
  StreamSubscription<location.LocationData>? _locationChangeStreamSubscription;

  /// Returns `true` if [listen] has already been executed.
  ///
  /// すでに[listen]が実行されている場合`true`を返します。
  bool get listening {
    if (!permitted) {
      return false;
    }
    if (_locationChangeStreamSubscription == null) {
      return false;
    }
    return true;
  }

  /// If permission is granted by executing [initialize], returns `true`.
  ///
  /// [initialize]を実行してパーミッションが許可されている場合は`true`を返します。
  bool get permitted => _permissionStatus == PermissionStatus.granted;
  PermissionStatus _permissionStatus = PermissionStatus.denied;

  /// Returns `true` if [initialize] is executed and permissions for background are allowed.
  ///
  /// [initialize]を実行してバックグラウンド向けのパーミッションが許可されている場合は`true`を返します。
  bool get permittedBackground =>
      _backgroundPermissionStatus == PermissionStatus.granted;
  PermissionStatus _backgroundPermissionStatus = PermissionStatus.denied;

  /// Initialization.
  ///
  /// Location information is available or not and permissions are granted.
  ///
  /// 初期化を行います。
  ///
  /// 位置情報が利用可能かどうかとパーミッションの許可を行います。
  Future<void> initialize({
    Duration timeout = const Duration(seconds: 60),
  }) async {
    if (_initializeCompleter != null) {
      return _initializeCompleter?.future;
    }
    _initializeCompleter = Completer<void>();
    try {
      if (!await _location.serviceEnabled().timeout(timeout)) {
        throw Exception(
          "Location service not available. The platform may not be supported or it may be disabled in the settings. please confirm.",
        );
      }
      _permissionStatus =
          await Permission.locationWhenInUse.status.timeout(timeout);
      if (_permissionStatus != PermissionStatus.granted) {
        _permissionStatus =
            await Permission.locationWhenInUse.request().timeout(timeout);
      }
      if (_permissionStatus != PermissionStatus.granted) {
        throw Exception(
          "You are not authorized to use the location information service. Check the permission settings.",
        );
      }
      if (adapter.enableBackgroundLocation) {
        _backgroundPermissionStatus =
            await Permission.locationAlways.status.timeout(timeout);
        if (_backgroundPermissionStatus != PermissionStatus.granted) {
          _backgroundPermissionStatus =
              await Permission.locationAlways.request().timeout(timeout);
        }
      }
      _initialized = true;
      notifyListeners();
      _initializeCompleter?.complete();
      _initializeCompleter = null;
    } catch (e) {
      _initializeCompleter?.completeError(e);
      _initializeCompleter = null;
      rethrow;
    } finally {
      _initializeCompleter?.complete();
      _initializeCompleter = null;
    }
  }

  /// Starts acquiring location information.
  ///
  /// The update interval can be specified by specifying [updateInterval].
  ///
  /// [LocationData] is stored in [value], so refer to it.
  ///
  /// You can also specify the accuracy with [accuracy] or [distanceFilterMeters].
  ///
  /// 位置情報の取得を開始します。
  ///
  /// [updateInterval]を指定することで更新間隔を指定できます。
  ///
  /// [value]に[LocationData]が格納されるのでそこを参照してください。
  ///
  /// [accuracy]や[distanceFilterMeters]で精度を指定することもできます。
  Future<void> listen({
    LocationAccuracy? accuracy,
    double? distanceFilterMeters,
    Duration updateInterval = const Duration(seconds: 1),
    Duration timeout = const Duration(seconds: 60),
  }) async {
    if (_listenCompleter != null) {
      return _listenCompleter?.future;
    }
    if (_updateInterval == updateInterval && listening) {
      return;
    }
    _listenCompleter = Completer<void>();
    _updateInterval = updateInterval;
    try {
      await initialize(timeout: timeout);
      await _location.changeSettings(
        accuracy: (accuracy ?? LocationMasamuneAdapter.primary.defaultAccuracy)
            .toGeoLocatorLocationAccuracy(),
        distanceFilter: distanceFilterMeters ??
            LocationMasamuneAdapter.primary.defaultDistanceFilterMeters,
      );
      if (adapter.enableBackgroundLocation && permittedBackground) {
        await _location.enableBackgroundMode(enable: true);
      }
      _value = (await _location.getLocation()).toLocationData();
      _locationChangeStreamSubscription?.cancel();
      _locationChangeStreamSubscription = _location.onLocationChanged.listen(
        (event) {
          _updated = true;
          _value = event.toLocationData();
        },
      );
      _updated = false;
      _timer?.cancel();
      _timer = Timer.periodic(_updateInterval, (timer) {
        if (!_updated) {
          return;
        }
        _updated = false;
        notifyListeners();
      });
      notifyListeners();
      _listenCompleter?.complete();
      _listenCompleter = null;
    } catch (e) {
      _listenCompleter?.completeError(e);
      _listenCompleter = null;
      rethrow;
    } finally {
      _listenCompleter?.complete();
      _listenCompleter = null;
    }
  }

  /// Cancel acquisition of [LocationData].
  ///
  /// [LocationData]の取得をキャンセルします。
  Future<void> unlisten() async {
    _updated = false;
    _timer?.cancel();
    _timer = null;
    _locationChangeStreamSubscription?.cancel();
    _locationChangeStreamSubscription = null;
  }

  @override
  void dispose() {
    unlisten();
    super.dispose();
  }
}

@immutable
class _$LocationQuery {
  const _$LocationQuery();

  @useResult
  _$_LocationQuery call() => _$_LocationQuery(
        hashCode.toString(),
      );
}

@immutable
class _$_LocationQuery extends ControllerQueryBase<Location> {
  const _$_LocationQuery(
    this._name,
  );

  final String _name;

  @override
  Location Function() call(Ref ref) {
    return () => Location();
  }

  @override
  String get queryName => _name;
  @override
  bool get autoDisposeWhenUnreferenced => false;
}

extension on LocationAccuracy {
  location.LocationAccuracy toGeoLocatorLocationAccuracy() {
    switch (this) {
      case LocationAccuracy.lowest:
        return location.LocationAccuracy.powerSave;
      case LocationAccuracy.low:
        return location.LocationAccuracy.low;
      case LocationAccuracy.medium:
        return location.LocationAccuracy.balanced;
      case LocationAccuracy.high:
        return location.LocationAccuracy.high;
      case LocationAccuracy.best:
        return location.LocationAccuracy.high;
      case LocationAccuracy.navigation:
        return location.LocationAccuracy.navigation;
      default:
        return location.LocationAccuracy.reduced;
    }
  }
}

extension on location.LocationData {
  LocationData toLocationData() {
    return LocationData(
      latitude: latitude ?? 0.0,
      longitude: longitude ?? 0.0,
      altitude: altitude,
      accuracy: accuracy,
      heading: heading,
      speed: speed,
      speedAccuracy: speedAccuracy,
      timestamp: time != null
          ? DateTime.fromMillisecondsSinceEpoch(time!.toInt())
          : null,
    );
  }
}
